/*
 *  TINET (TCP/IP Protocol Stack)
 * 
 *  Copyright (C) 2001-2009 by Dep. of Computer Science and Engineering
 *                   Tomakomai National College of Technology, JAPAN
 *  Copyright (C) 2014 Cores Co., Ltd. Japan
 * 
 *  L쌠҂́Cȉ(1)`(4)̏𖞂ꍇɌC{\tgEF
 *  Ai{\tgEFAς̂܂ށDȉjgpEE
 *  ρEĔzziȉCpƌĂԁj邱Ƃ𖳏ŋD
 *  (1) {\tgEFA\[XR[ȟ`ŗpꍇɂ́CL̒
 *      \C̗pщL̖ۏ؋K肪Ĉ܂܂̌`Ń\[
 *      XR[hɊ܂܂Ă邱ƁD
 *  (2) {\tgEFACCu`ȂǁC̃\tgEFAJɎg
 *      pł`ōĔzzꍇɂ́CĔzzɔhLgip
 *      ҃}jAȂǁjɁCL̒쌠\C̗pщL
 *      ̖ۏ؋Kfڂ邱ƁD
 *  (3) {\tgEFAC@ɑgݍނȂǁC̃\tgEFAJɎg
 *      płȂ`ōĔzzꍇɂ́Ĉꂩ̏𖞂
 *      ƁD
 *    (a) ĔzzɔhLgip҃}jAȂǁjɁCL̒
 *        쌠\C̗pщL̖ۏ؋Kfڂ邱ƁD
 *    (b) Ĕzž`ԂCʂɒ߂@ɂāCTOPPERSvWFNg
 *        񍐂邱ƁD
 *  (4) {\tgEFA̗pɂ蒼ړI܂͊ԐړIɐ邢Ȃ鑹
 *      QCL쌠҂TOPPERSvWFNgƐӂ邱ƁD
 *      ܂C{\tgEFÃ[U܂̓Gh[ÛȂ闝
 *      RɊÂCL쌠҂TOPPERSvWFNg
 *      Ɛӂ邱ƁD
 * 
 *  {\tgEFÁCۏ؂Œ񋟂Ă̂łDL쌠҂
 *  TOPPERSvWFNǵC{\tgEFAɊւāC̎gpړI
 *  ɑ΂K܂߂āCȂۏ؂sȂD܂C{\tgEF
 *  A̗pɂ蒼ړI܂͊ԐړIɐȂ鑹QɊւĂC
 *  ̐ӔC𕉂ȂD
 * 
 *  @(#) $Id: ether_phy.c 327 2017-11-27 04:24:02Z coas-nagasima $
 */

#include "ether_phy.h"
#include <kernel.h>
#include <sil.h>

#define PHY_CONTROL_REGISTER	0x00

#define PHY_CONTROL_RESET			0x8000
#define PHY_CONTROL_LOOPBACK		0x4000
#define PHY_CONTROL_SPEED_SELECT	0x2000
#define PHY_CONTROL_A_N_ENABLE		0x1000
#define PHY_CONTROL_POWER_DOWN		0x0800
#define PHY_CONTROL_ISOLATE			0x0400
#define PHY_CONTROL_RESTART_A_N		0x0200
#define PHY_CONTROL_DUPLEX_MODE		0x0100
#define PHY_CONTROL_COLLISION_TEST	0x0080

#define PHY_STATUS_REGISTER		0x01

#define PHY_STATUS_100BASE_T4				0x8000
#define PHY_STATUS_100BASE_TX_FULL_DUPLEX	0x4000
#define PHY_STATUS_100BASE_TX_HALF_DUPLEX	0x2000
#define PHY_STATUS_10BASE_T_FULL_DUPLEX		0x1000
#define PHY_STATUS_10BASE_T_HALF_DUPLEX		0x0800
#define PHY_STATUS_A_N_COMPLETE				0x0020
#define PHY_STATUS_REMOTE_FAULT				0x0010
#define PHY_STATUS_A_N_ABILITY				0x0008
#define PHY_STATUS_LINK_STATUS				0x0004
#define PHY_STATUS_JABBER_DETECT			0x0002
#define PHY_STATUS_EXTENDED_CAPABILITY		0x0001

#define PHY_AN_ADVERTISEMENT_REGISTER	0x04

#define PHY_AN_AD_100BASE_T4				0x0200
#define PHY_AN_AD_100BASE_TX_FULL_DUPLEX	0x0100
#define PHY_AN_AD_100BASE_TX_HALF_DUPLEX	0x0080
#define PHY_AN_AD_10BASE_T_FULL_DUPLEX		0x0040
#define PHY_AN_AD_10BASE_T_HALF_DUPLEX		0x0020
#define PHY_AN_AD_SELECTOR_FIELD			0x0001

#define PHY_AN_STATUS_REGISTER	0x05

#define PHY_AN_STATUS_100BASE_T4				0x0200
#define PHY_AN_STATUS_100BASE_TX_FULL_DUPLEX	0x0100
#define PHY_AN_STATUS_100BASE_TX_HALF_DUPLEX	0x0080
#define PHY_AN_STATUS_10BASE_T_FULL_DUPLEX		0x0040
#define PHY_AN_STATUS_10BASE_T_HALF_DUPLEX		0x0020

static void phy_write_bit(bool_t bit);
static void phy_release_bus();
static bool_t phy_read_bit();
static void phy_single_rel_bus();

PHY_STATE_T phy_reset(PHY_STATE_T state, uint8_t phy_addr)
{
	switch(state){
	case PHY_STATE_UNINIT:
		/* PHỸZbg */
		phy_write_reg(phy_addr, PHY_CONTROL_REGISTER, PHY_CONTROL_RESET);

		/* 500ms҂ */
		/* tslp_tsk(500); svH */
		/* continue; */
	case PHY_STATE_RESETING:
		/* Zbg̊mF */
		if((phy_read_reg(phy_addr, PHY_CONTROL_REGISTER) & PHY_CONTROL_RESET) != 0)
			return PHY_STATE_RESETING;

		return PHY_STATE_RESET;
	default:
		return state;
	}
}

PHY_STATE_T phy_initialize(PHY_STATE_T state, uint8_t phy_addr, PHY_MODE_T *mode)
{
#ifndef __RX
	static int reset = 0;
#endif
	uint16_t status;
	*mode = (PHY_MODE_T)0;

	switch(state){
	case PHY_STATE_RESET:
		/* I[glSVG[V̂m点Lɐݒ */
		phy_write_reg(phy_addr, PHY_AN_ADVERTISEMENT_REGISTER, 
			PHY_AN_AD_100BASE_T4 | PHY_AN_AD_100BASE_TX_FULL_DUPLEX | PHY_AN_AD_100BASE_TX_HALF_DUPLEX
			| PHY_AN_AD_10BASE_T_FULL_DUPLEX | PHY_AN_AD_10BASE_T_HALF_DUPLEX | PHY_AN_AD_SELECTOR_FIELD);

		/* I[glSVG[VLɐݒ */
		phy_write_reg(phy_addr, PHY_CONTROL_REGISTER, PHY_CONTROL_SPEED_SELECT | PHY_CONTROL_A_N_ENABLE);

		/* continue; */
	case PHY_STATE_LINKE_WAIT:
#ifndef __RX
		reset++;
		if(reset < 2){
#endif
		/* N̊mF */
		if(!phy_is_link(phy_addr))
			return PHY_STATE_LINKE_WAIT;
#ifndef __RX
		}
		reset = 0;
#endif
		/* continue; */
	case PHY_STATE_LINKED:
#ifndef __RX
		reset++;
		if(reset < 2){
#endif
		/* I[glSVG[V̊mF */
		if((phy_read_reg(phy_addr, PHY_STATUS_REGISTER) & PHY_STATUS_A_N_COMPLETE) == 0)
			return PHY_STATE_LINKED;
#ifndef __RX
		}
#endif
		status = phy_read_reg(phy_addr, PHY_AN_STATUS_REGISTER);

		if ((status & (PHY_AN_STATUS_100BASE_T4 | PHY_AN_STATUS_10BASE_T_FULL_DUPLEX | PHY_AN_STATUS_100BASE_TX_FULL_DUPLEX)) != 0)
			*mode = (PHY_MODE_T)(((int)*mode) | 0x01);

		if ((status & (PHY_AN_STATUS_100BASE_T4 | PHY_AN_STATUS_100BASE_TX_FULL_DUPLEX | PHY_AN_STATUS_100BASE_TX_HALF_DUPLEX)) != 0)
			*mode = (PHY_MODE_T)(((int)*mode) | 0x02);

		return PHY_STATE_NEGOTIATED;
	}

	return state;
}

bool_t phy_is_link(uint8_t phy_addr)
{
#ifdef __RX
	/* N̊mF */
	return (phy_read_reg(phy_addr, PHY_STATUS_REGISTER) & PHY_STATUS_LINK_STATUS) != 0;
#else
	return true;
#endif
}

uint16_t phy_read_reg(uint8_t phy_addr, uint8_t reg_addr)
{
	uint16_t result = 0, bit;
	int i;

	/* PRE F32̘A1b */
	for (i = 0; i < 32; i++)
		phy_write_bit(true);

	/* ST Ft[̐擪01b̃Cg */
	phy_write_bit(false);
	phy_write_bit(true);

	/* OP FANZXʂR[h̃Cg */
	phy_write_bit(true);
	phy_write_bit(false);

	/* PHYAD FPHY-LSĨAhX1Ԃ̏ꍇA00001bCgiMSB 珇CgjB */
	phy_write_bit((phy_addr & 0x10) != 0);
	phy_write_bit((phy_addr & 0x08) != 0);
	phy_write_bit((phy_addr & 0x04) != 0);
	phy_write_bit((phy_addr & 0x02) != 0);
	phy_write_bit((phy_addr & 0x01) != 0);

	/* REGAD FWX^AhX1Ԃ̏ꍇA00001bCgiMSB珇CgjB */
	phy_write_bit((reg_addr & 0x10) != 0);
	phy_write_bit((reg_addr & 0x08) != 0);
	phy_write_bit((reg_addr & 0x04) != 0);
	phy_write_bit((reg_addr & 0x02) != 0);
	phy_write_bit((reg_addr & 0x01) != 0);

	/* TA FMII/RMIIC^tF[XŃf[^̑M؂ւ鎞 */
	/* uoXviZ0ƕ\Ljs */
	phy_release_bus();

	/* DATA F16rbg̃f[^BMSB珇[h */
	for (bit = 0x8000; bit != 0; bit >>= 1) {
		if(phy_read_bit())
			result |= bit;
	}

	/* IDLE FMIIǗtH[}bg͂܂ł̑ҋ@ */
	/* łTAɃoXς݂ł萧sv */
	phy_release_bus();

	return result;
}

void phy_write_reg(uint8_t phy_addr, uint8_t reg_addr, uint16_t reg_data)
{
	int i;
	uint16_t bit;

	/* PRE F32̘A1b */
	for (i = 0; i < 32; i++)
		phy_write_bit(true);

	/* ST Ft[̐擪01b̃Cg */
	phy_write_bit(false);
	phy_write_bit(true);

	/* OP FANZXʂR[h̃Cg */
	phy_write_bit(false);
	phy_write_bit(true);

	/* PHYAD FPHY-LSĨAhX1Ԃ̏ꍇA00001bCgiMSB 珇CgjB */
	phy_write_bit((phy_addr & 0x10) != 0);
	phy_write_bit((phy_addr & 0x08) != 0);
	phy_write_bit((phy_addr & 0x04) != 0);
	phy_write_bit((phy_addr & 0x02) != 0);
	phy_write_bit((phy_addr & 0x01) != 0);

	/* REGAD FWX^AhX1Ԃ̏ꍇA00001bCgiMSB珇CgjB */
	phy_write_bit((reg_addr & 0x10) != 0);
	phy_write_bit((reg_addr & 0x08) != 0);
	phy_write_bit((reg_addr & 0x04) != 0);
	phy_write_bit((reg_addr & 0x02) != 0);
	phy_write_bit((reg_addr & 0x01) != 0);

	/* TA FMII/RMIIC^tF[XŃf[^̑M؂ւ鎞 */
	/* 10bCg */
	phy_write_bit(true);
	phy_write_bit(false);

	/* DATA F16rbg̃f[^BMSB珇Cg */
	for (bit = 0x8000; bit != 0; bit >>= 1) {
		phy_write_bit((reg_data & bit) != 0);
	}

	/* IDLE FMIIǗtH[}bg͂܂ł̑ҋ@ */
	/* uPƃoXviXƕ\Ljs */
	phy_single_rel_bus();
}

#define ETHERC_PIR_WAIT 4

/*
 * 1 rbgf[^̃Cg
 */
static void phy_write_bit(bool_t bit)
{
	int i;
	uint32_t data = bit ? ETHERC_PIR_MDO : 0;

	/* PHYC^tF[XWX^ւ̃Cg */
	sil_wrw_mem(ETHERC_PIR, data | ETHERC_PIR_MMD | 0/*ETHERC_PIR_MDC*/);

	for(i = ETHERC_PIR_WAIT; i > 0; i--);

	/* PHYC^tF[XWX^ւ̃Cg */
	sil_wrw_mem(ETHERC_PIR, data | ETHERC_PIR_MMD | ETHERC_PIR_MDC);

	for(i = 2 * ETHERC_PIR_WAIT; i > 0; i--);

	/* PHYC^tF[XWX^ւ̃Cg */
	sil_wrw_mem(ETHERC_PIR, data | ETHERC_PIR_MMD | 0/*ETHERC_PIR_MDC*/);

	for(i = ETHERC_PIR_WAIT; i > 0; i--);
}

/*
 * oX
 */
static void phy_release_bus()
{
	int i;

	/* PHYC^tF[XWX^ւ̃Cg */
	sil_wrw_mem(ETHERC_PIR, 0/*ETHERC_PIR_MMD*/ | 0/*ETHERC_PIR_MDC*/);

	for(i = ETHERC_PIR_WAIT; i > 0; i--);

	/* PHYC^tF[XWX^ւ̃Cg */
	sil_wrw_mem(ETHERC_PIR, 0/*ETHERC_PIR_MMD*/ | ETHERC_PIR_MDC);

	for(i = 2 * ETHERC_PIR_WAIT; i > 0; i--);

	/* PHYC^tF[XWX^ւ̃Cg */
	sil_wrw_mem(ETHERC_PIR, 0/*ETHERC_PIR_MMD*/ | 0/*ETHERC_PIR_MDC*/);

	for(i = ETHERC_PIR_WAIT; i > 0; i--);
}

/*
 * 1 rbgf[^̃[h
 */
static bool_t phy_read_bit()
{
	bool_t bit;
	int i;

	for(i = ETHERC_PIR_WAIT; i > 0; i--);

	/* PHYC^tF[XWX^ւ̃Cg */
	sil_wrw_mem(ETHERC_PIR, 0/*ETHERC_PIR_MMD*/ | ETHERC_PIR_MDC);

	for(i = ETHERC_PIR_WAIT; i > 0; i--);

	/* PHYC^tF[XWX^ւ̃Cg */
	bit = (sil_rew_mem(ETHERC_PIR) & ETHERC_PIR_MDI) != 0;

	for(i = ETHERC_PIR_WAIT; i > 0; i--);

	/* PHYC^tF[XWX^ւ̃Cg */
	sil_wrw_mem(ETHERC_PIR, 0/*ETHERC_PIR_MMD*/ | 0/*ETHERC_PIR_MDC*/);

	for(i = ETHERC_PIR_WAIT; i > 0; i--);

	return bit;
}

/*
 * PƃoX
 */
static void phy_single_rel_bus()
{
	int i;

	for(i = ETHERC_PIR_WAIT; i > 0; i--);

	/* PHYC^tF[XWX^ւ̃Cg */
	sil_wrw_mem(ETHERC_PIR, ETHERC_PIR_MMD | ETHERC_PIR_MDC);

	for(i = 2 * ETHERC_PIR_WAIT; i > 0; i--);

	/* PHYC^tF[XWX^ւ̃Cg */
	sil_wrw_mem(ETHERC_PIR, 0/*ETHERC_PIR_MMD*/ | 0/*ETHERC_PIR_MDC*/);

	for(i = ETHERC_PIR_WAIT; i > 0; i--);
}
